/***************************************************************************
                     ascii.cpp  -  ASCII file data source
                             -------------------
    begin                : Fri Oct 17 2003
    copyright            : (C) 2003 The University of Toronto
    email                :
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ascii.h"
#include "asciiconfig.h"

#include <kdebug.h>

#include <qfile.h>
#include <qlayout.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qradiobutton.h>
#include <qcheckbox.h>
#include <qregexp.h>
#include <qstring.h>
#include <qstringlist.h>

#include <assert.h>
#include <ctype.h>
#include <stdlib.h>


AsciiSource::AsciiSource(KConfig *cfg, const QString& filename, const QString& type)
: KstDataSource(cfg, filename, type) {
  if (!type.isEmpty() && type != "ASCII") {
    return;
  }

 // Create all useful private variables from the config file,
  cfg->setGroup("ASCII General");
  _commentDelimiters = QRegExp::escape(cfg->readEntry("Comment Delimiters", "#/c!;"));
  _dataStartLine = cfg->readNumEntry("Header lines", 0);
  _useVarNames = cfg->readBoolEntry("Use variables", false);
  _useUnits = cfg->readBoolEntry("Use units", false);
  _unitsLine = cfg->readNumEntry("Units line", 4);
  _useFixedWidth = cfg->readBoolEntry("Fixed width", false);
  _columnWidth = cfg->readNumEntry("Column width", 13);
  _useCustomDelimiter = cfg->readBoolEntry("Custom delimiter", false);
  _variableDelimiter = QRegExp::escape(cfg->readEntry("Variable delimiter", ";"));

  // Other inits
  _filename = filename;
  _byteLength = 0;
  _rowIndex = 0L;
  
  _fieldList.clear();
  _fieldList = fieldListFor(filename, *cfg);
  
  // Initialize _rowIndex
  if (!initRowIndex()) {
    _valid = false;
  }
   
  update();
}


AsciiSource::~AsciiSource() {
  if (_rowIndex) {
    free(_rowIndex);
    _rowIndex = 0L;
    _numLinesAlloc = 0;
  }
}


bool AsciiSource::initRowIndex() {
  // First, clean up
  if (_rowIndex) {
    free(_rowIndex);
    _rowIndex = 0L;
    _numLinesAlloc = 0;
  }
  // Reallocate
  _rowIndex = (int *)malloc(32768 * sizeof(int));
  _numLinesAlloc = 32768;
  // Prepare for parsing
  int offset = 0;
  QFile file(_filename);
  if (!file.open(IO_ReadOnly)) {
    return false;
  }
  QTextStream TS(&file);
  QString line;
  QRegExp commentRx(QString("^\\s*[%1].*").arg(_commentDelimiters));
  // Skip header
  for (int i = 0; i < _dataStartLine - 1; i++) {
    line = TS.readLine();
  }
  // Find the first data line that is not comments
  bool done = false;
  while (!TS.atEnd() && !done) {
    // Store position before moving cursor one line further
    offset = TS.device()->at();
    // Read next line to see whether it's a comment
    line = TS.readLine();
    if (!line.isEmpty() && !commentRx.exactMatch(line)) {
      done = true;
      _rowIndex[0] = offset;
      // kdDebug() << "Found first data line at: " << offset << endl;
    }
  }
  file.close();
  
  if (done) { 
    return true;
  }
  else return false;
}


#define MAXBUFREADLEN 32768
KstObject::UpdateType AsciiSource::update(int u) {
  // This method checks for new data (i.e. a bigger file with new data lines).
  // The beginning of each data line ("frame") is stored in _rowIndex, which is used 
  // to read the data in readField().
  
  Q_UNUSED(u)
  bool forceUpdate = false;
  char tmpbuf[MAXBUFREADLEN];
  QFile file(_filename);

  if (!_valid) {
    _numFrames = 0;
    _byteLength = 0;
    initRowIndex();
  }

  // If the list of field changes it means we have to reset _rowIndex 
  // and rebuild it from scratch as byte offsets will all change !
  // If we want to handle files where data is not only appended to a fixed
  // list of columns but columns may (dis)appear, we have to reset _numFrames to 0
  // and rescan everything ! Could be done at this point...
  
  // _fields.clear(); // FIXME: only clear if the fields have changed

  if (file.exists()) {
    if (uint(_byteLength) != file.size() || !_valid) {
      forceUpdate = true;
    }
    _byteLength = file.size();
  } else {
    _valid = false;
    return KstObject::NO_CHANGE;
  }

  if (!file.open(IO_ReadOnly)) {
    // quietly fail - no data to be had here
    _valid = false;
    return KstObject::NO_CHANGE;
  }

  _valid = true;

  int bufstart, bufread;
  bool new_data = false;
  do {
    /* Read the tmpbuffer, starting at row_index[_numFrames] */
    if (_byteLength - _rowIndex[_numFrames] > MAXBUFREADLEN) {
      bufread = MAXBUFREADLEN;
    } else {
      bufread = _byteLength - _rowIndex[_numFrames];
    }

    bufstart = _rowIndex[_numFrames];
    file.at(bufstart); // expensive?
    file.readBlock(tmpbuf, bufread);

    bool is_comment = false, has_dat = false;
    for (int i = 0; i < bufread; i++) {
      // Statistically numbers should be more common than comments, so let's
      // check for them first
      // if isdigit() generates a function call, it would be better to write
      // this inline, but I'm not convinced that compilers are that stupid
      // yet.  Profiling demo.kst shows this code path is hit > 10^5 times.
      if (isdigit(tmpbuf[i])) {
        if (!is_comment) {
          has_dat = true;
        }
      } else if (tmpbuf[i] == '\n' || tmpbuf[i] == '\r') {
        if (has_dat) {
          _numFrames++;
          if (_numFrames >= _numLinesAlloc) {
            _numLinesAlloc += 32768;
            _rowIndex = (int *)realloc(_rowIndex, _numLinesAlloc*sizeof(int));
          }
          new_data = true;
        }
        _rowIndex[_numFrames] = bufstart + i + 1;
        has_dat = is_comment = false;
      } else if (_commentDelimiters.contains(tmpbuf[i])) {
        is_comment = true;
      }
    }
  } while (bufread == MAXBUFREADLEN);

  file.close();

  return forceUpdate ? KstObject::UPDATE : (new_data ? KstObject::UPDATE : KstObject::NO_CHANGE);
}


int AsciiSource::readField(double *v, const QString& field, int s, int n) {
  if (n < 0) {
    n = 1; /* n < 0 means read one sample, not frame - irrelevant here */
  }

  if (field == "INDEX") {
    for (int i = 0; i < n; i++) {
      v[i] = double(s + i);
    }
    return n;
  }

  if (!isValidField(field)) {
    return 0;
  }

  // Prepare a char var to speed up the test below, otherwise we could just use _variableDelimiter.contains()
  char delim_ascii[] = " ";
  if (_useCustomDelimiter) {
    _variableDelimiter.truncate(1); 
    strncpy(delim_ascii, _variableDelimiter.ascii(),2);
  }
  
  int col =  (_useVarNames) ? _fieldList.findIndex(field) : field.toInt();
  int bufstart = _rowIndex[s];
  int bufread = _rowIndex[s + n] - bufstart;
  
  if (bufread <= 0) {
   return 0;
  }

  QFile file(_filename);
  if (!file.open(IO_ReadOnly)) {
    _valid = false;
    return 0;
  }

  char *tmpbuf = new char[bufread];

  file.at(bufstart);
  file.readBlock(tmpbuf, bufread);

  // FIXME: I don't see why this was changed to memset() instead of just
  //        setting the required elements to 0 inline when required, as it
  //        did before.
  memset( v, 0, n * sizeof( double ) );

  if (_useFixedWidth) { // Fixed width columns : we can access the right bytes directly
    for (int i = 0; i < n; i++, s++) {
      // Read appropriate column and convert to double
      v[i] = (double) atof(tmpbuf + _rowIndex[i] - _rowIndex[0] + _columnWidth*(col-1)); 
    }
  }
  else { // read almost like in the usual way, only with custom delimiter added
    for (int i = 0; i < n; i++, s++) {
      bool done = false;
      bool incol = false;
      int i_col = 0;
      for (int ch = _rowIndex[s] - bufstart; !done && ch < bufread; ch++) {
	if (isspace(tmpbuf[ch]) || tmpbuf[ch]==delim_ascii[0]) {
          if (tmpbuf[ch] == '\n' || tmpbuf[ch] == '\r' ) {
            done = true;
          } else {
            incol = false;
          }
        } else if (_commentDelimiters.contains(tmpbuf[ch])) {  
            // Do we really want this ???
	    // This impacts performance and is probably useless
	    // Furthermore, it makes some values impossible to access:
	    // if e.g. column 3 has values like '####' in frames 10-20, then you will never 
	    // be able to read the corresponding values for column numbers > 3 !
	    // I'd favor removing this test, but not until I get some feedback...
	    done = true;
        } else {
          if (!incol) {
            incol = true;
            i_col++;
            if (i_col == col) {
              done = true;
              v[i] = atof(tmpbuf + ch);
            }
          }
        }
      }
    }
  }

  delete[] tmpbuf;

  file.close();

  return n;
}


bool AsciiSource::isValidField(const QString& field) const {
  return _fieldList.contains(field);
}


int AsciiSource::samplesPerFrame(const QString &field) {
  Q_UNUSED(field)
  return 1;
}


int AsciiSource::frameCount(const QString& field) const {
  Q_UNUSED(field)
  return _numFrames;
}


QString AsciiSource::fileType() const {
  return "ASCII";
}


bool AsciiSource::isEmpty() const {
  return _numFrames < 1;
}


/** Builds the list of fields for a given file. Many cases here :
    1) variable names are given in the file: simply read them
    2) no variable names given: go to the first data line to automatically detect the 
       number of columns and use column numbers as names. **/
static QStringList fieldListFor(const QString& filename, const KConfig& cfg) {
  QStringList rc;
  QFile file(filename);
  QTextStream TS(&file);
  QString line, variables, units;
  QRegExp del;
  int i = 0;
  
  if (!file.open(IO_ReadOnly)) {
      return rc;
  }
  
  rc += "INDEX";
  
  // Store config info
  int dataStartLine = cfg.readNumEntry("Header lines", 0);
  bool useVarNames = cfg.readBoolEntry("Use variables", false);
  int varNamesLine = cfg.readNumEntry("Variable names line", 3);
  bool useUnits = cfg.readBoolEntry("Use units", false);
  int unitsLine = cfg.readNumEntry("Units line", 4);
  bool useCustomDelimiter = cfg.readBoolEntry("Custom delimiter", false);
  bool useFixedWidth = cfg.readBoolEntry("Fixed width", false);
  int columnWidth = cfg.readNumEntry("Column width", 13);
  QString variableDelimiter = cfg.readEntry("Variable delimiter", ";");
  QString commentDelimiters = QRegExp::escape(cfg.readEntry("Comment Delimiters", "#/c!;"));
  
  // Read through header  (if any, otherwise the loop is not entered)
  while (i < dataStartLine) {
    line = TS.readLine(); 
    i++;
    if (i == varNamesLine && useVarNames) { variables = line; }
    if (i == unitsLine && useUnits) { units = line; }
  }
  
  // Prepare regexps
  QRegExp varRx( "\\t|(\\s)+|(\\s)*\"(\\s)*" );
  if (useCustomDelimiter) { varRx.setPattern(varRx.pattern() + "|" + variableDelimiter); }
  QRegExp commentRx(QString("^\\s*[%1].*").arg(commentDelimiters));
  
  // First case: field names are given in the file
  QString var("");
  QString unit("");
  QStringList tmpVars;
  QStringList tmpUnits;
  uint charNumber = variables.length();
  if (useUnits && (units.length() > charNumber)) {
    charNumber = units.length();
  }
  if (useVarNames) {  
    if (useFixedWidth) {
      for (uint j = 0; j < charNumber; j++) {
	if (useUnits) { // Read units if required
	  if (!units[j].isNull()) {
	    unit += units[j];
	  }
	  if ((j+1)%columnWidth == 0) {
	    tmpUnits += unit.stripWhiteSpace();
	    unit = "";
	  }
	}
	// Read variables
	var += variables[j];
	if ((j+1)%columnWidth == 0) {
	  tmpVars += var.stripWhiteSpace();
	  var = "";
	} 
      }
    }
    else { // Easier : just split with a QRexExp
      tmpVars += QStringList::split ( varRx, variables );  
      tmpUnits += QStringList::split ( varRx, units );  
    }
    // Now that we have the list of field names, append units if required
    if (useUnits) { 
      QStringList::Iterator it1, it2;
      for ( it1 = tmpVars.begin(), it2 = tmpUnits.begin();
              it1 != tmpVars.end() && it2 != tmpUnits.end();
	      ++it1, ++it2 ) {
        *it1 += QString(" (%1)").arg(*it2);
      }
    }
    
    // Store the final list
    rc += tmpVars;
  }
  
  else { // Fall back to column numbers
    // Find the first data line that is not comments
    bool done = false;
    while (!TS.atEnd() && !done) {
        line = TS.readLine();
        if (!line.isEmpty() && !commentRx.exactMatch(line)) {
          done = true;
        }
    } 
    // Now, build field list
    int cnt = QStringList::split( varRx, line ).count();
    for (int i = 1; i <= cnt; ++i) {
      rc += QString::number(i);
    }
  }
  
  file.close();

  return rc;
}


// QStringList AsciiSource::fieldList() const: removed here since basic implementation is OK
// as it returns _fieldList which is now cached


void AsciiSource::save(QTextStream &ts, const QString& indent) {
  KstDataSource::save(ts, indent);
}


class ConfigWidgetAscii : public KstDataSourceConfigWidget {
  public:
    ConfigWidgetAscii() : KstDataSourceConfigWidget() {
      QGridLayout *layout = new QGridLayout(this, 1, 1);
      _ac = new AsciiConfig(this);
      layout->addWidget(_ac, 0, 0);
      layout->activate();
    }

    virtual ~ConfigWidgetAscii() {}

    virtual void setConfig(KConfig *cfg) {
      KstDataSourceConfigWidget::setConfig(cfg);
      cfg->setGroup("ASCII General");
      _ac->_commentDelimiters->setText(cfg->readEntry("Comment Delimiters", "#/c!;"));
      _ac->_filenamePattern->setText(cfg->readEntry("Filename pattern", "*.dat"));
      _ac->_dataStartLine->setValue(cfg->readNumEntry("Header lines", 0));
      _ac->_useVarNames->setChecked(cfg->readBoolEntry("Use variables", false));
      _ac->_varNamesLine->setValue(cfg->readNumEntry("Variable names line", 3));
      _ac->_useUnits->setChecked(cfg->readBoolEntry("Use units", false));
      _ac->_unitsLine->setValue(cfg->readNumEntry("Units line", 4));
      _ac->_freeFormat->setChecked(cfg->readBoolEntry("Free format", true));
      _ac->_fixedWidth->setChecked(cfg->readBoolEntry("Fixed width", false));
      _ac->_columnWidth->setValue(cfg->readNumEntry("Column width", 13));
      _ac->_customDelimiter->setChecked(cfg->readBoolEntry("Custom delimiter", false));
      _ac->_variableDelimiter->setText(cfg->readEntry("Variable delimiter", ";"));
    }

    virtual void save() {
      assert(_cfg);
      _cfg->setGroup("ASCII General");
      _cfg->writeEntry("Comment Delimiters", _ac->_commentDelimiters->text());
      _cfg->writeEntry("Filename pattern", _ac->_filenamePattern->text());
      _cfg->writeEntry("Header lines", _ac->_dataStartLine->value());
      _cfg->writeEntry("Use variables", _ac->_useVarNames->isChecked());
      _cfg->writeEntry("Variable names line", _ac->_varNamesLine->value());
      _cfg->writeEntry("Use units", _ac->_useUnits->isChecked());
      _cfg->writeEntry("Units line", _ac->_unitsLine->value());
      _cfg->writeEntry("Free format", _ac->_freeFormat->isChecked());
      _cfg->writeEntry("Fixed width", _ac->_fixedWidth->isChecked());
      _cfg->writeEntry("Column width", _ac->_columnWidth->value());
      _cfg->writeEntry("Custom delimiter", _ac->_customDelimiter->isChecked());
      _cfg->writeEntry("Variable delimiter", _ac->_variableDelimiter->text());
    }

    AsciiConfig *_ac;
};


extern "C" {
KstDataSource *create_ascii(KConfig *cfg, const QString& filename, const QString& type) {
  return new AsciiSource(cfg, filename, type);
}

QStringList provides_ascii() {
  QStringList rc;
  rc += "ASCII";
  return rc;
}

/** understands_ascii: returns true if:
    1) the filename matches a user-given pattern
    2) or if the first line in the data block that is not a comment does start with a number **/
int understands_ascii(KConfig *cfg, const QString& filename) {
  QFile f(filename);

  cfg->setGroup("ASCII General");
  
  // First check by filename (the user should know what he's doing, so we can be pretty confident here)
  QRegExp filenamePattern(cfg->readEntry("Filename pattern", "*.dat"));
  filenamePattern.setWildcard(true);

  if ( filenamePattern.exactMatch(filename) ) { return 90; }

  // Otherwise, look inside the file, skipping the header and comments
  // Use a QTextStream here as we should never need to read very much data,
  // so that speed shouldn't be a factor
  QString delimiters = QRegExp::escape(cfg->readEntry("Comment Delimiters", "#/c!;"));
  int dataStartLine = cfg->readNumEntry("Header lines", 0);
  if (f.open(IO_ReadOnly)) {
    QString s;
    QTextStream TS(&f);

      // Skip header
      for (int j = 0; j <= dataStartLine; j++) { 
         s = TS.readLine();
      }
      // Look for first line in data block that is not a comment or empty
      while (QRegExp(QString("^\\s*[%1].*").arg(delimiters)).exactMatch(s) || s.isEmpty()==0 ) { 
         // comment: ignore and read next
	 s = TS.readLine();
	 if (s.isNull()) break; // do not get trapped if there is no data
      }
        
      // Check that this looks like data
      if (QRegExp("^[\\s]*(([Nn][Aa][Nn]|(\\-\\+)?[Ii][Nn][Ff]|[0-9\\+\\-\\.eE]+)[\\s]*)+").exactMatch(s)) {
        // a number - this may be an ascii file - assume that it is
        return 75; // What is the purpose of the test below ??? Looks wrong to me (NB) !
	// return QFile::exists(s.stripWhiteSpace()) ? 49 : 75;
      } else {
        return 20;
      }
  }
  else return 0;
}


QStringList fieldList_ascii(KConfig *cfg, const QString& filename, const QString& type, QString *typeSuggestion, bool *complete) {
  if ((!type.isEmpty() && !provides_ascii().contains(type)) ||
      !understands_ascii(cfg, filename)) {
    return QStringList();
  }

  if (complete) {
    *complete = true;
  }

  if (typeSuggestion) {
    *typeSuggestion = "ASCII";
  }

  return fieldListFor(filename, *cfg);
}


QWidget *widget_ascii(const QString& filename) {
  Q_UNUSED(filename)
  return new ConfigWidgetAscii;
}

}

// vim: ts=2 sw=2 et


